// ================================================================
//
// CBP_Pioneer3_Gripper_2.c
// 19-11-20 DEVEL
//
// Simple Pickup and put down
//
// =================================================================

#include <webots/display.h>
#include <webots/lidar.h>
#include <webots/motor.h>
#include <webots/robot.h>
#include <webots/light_sensor.h>
#include <webots/distance_sensor.h>

#include <math.h>
#include <stdio.h>
#include <stdlib.h>

# include "CBP_Plotter_2.h"

#define TIME_STEP 16
#define MAX_SPEED 6.28/3
#define PI 3.14159265358979323846
#define RANGE_THRESHOLD 2.0
#define GRIPPER_MOTOR_MAX_SPEED 0.1

#define STATE_FIND_OBJECT 1
#define STATE_PICKUP_OBJECT 2
#define STATE_PUTDOWN_OBJECT 3
#define STATE_FIND_LIGHT 4
#define STATE_DONE 5

#define DEBUG false

WbDeviceTag lidar;
WbDeviceTag left_wheel;
WbDeviceTag right_wheel;
WbDeviceTag display;
WbDeviceTag comp;
WbDeviceTag leftLight;
WbDeviceTag rightLight;
WbImageRef background;
WbDeviceTag USleft3;
WbDeviceTag USright4;
WbDeviceTag leftLightBack;
WbDeviceTag rightLightBack;

static WbDeviceTag gripper_motors[3];

int lidar_width;
static int px[180];

// attractive and repulsive factors
double k_att = 50.0;
double k_rep = 0.5;

// maximum distance of influence for the obstacles
double d_0 = 1.0;
static int time_step = 0;

// Display SICK laser data
static void update_display(WbDeviceTag display, WbImageRef background,float *scanvalue, int laser_width) {
  int half_width2 = laser_width / 2;
  int half_height2 = laser_width / 2;
  // (re)paint background
  wb_display_image_paste(display, background, 0, 0, false);
  // draw polygon scannings
  int py[laser_width];
  int i;
  for (i = 0; i < laser_width; i++) {
    if (scanvalue[i] > 8)  // range up to 8metres
      scanvalue[i] = 8;
    px[i] = half_width2 - (cos(i * PI / 180)) * ((90 * scanvalue[i]) / 8);
    py[i] = half_height2 - (sin(i * PI / 180)) * ((90 * scanvalue[i]) / 8);
  }
  // update Scan Display
  wb_display_draw_polygon(display, px, py, 180);
  
}

// Attractive force towards object dOBj
static double compute_force_obstacle(double dObj) {
  double force = 0;  
  if (dObj < d_0) {
    force = k_rep / pow(dObj,2);
  }  
  return force;
}

void init() {

  // get devices
  lidar = wb_robot_get_device("Sick LMS 291");
  left_wheel = wb_robot_get_device("left wheel");
  right_wheel = wb_robot_get_device("right wheel");
  display = wb_robot_get_device("display");
  comp = wb_robot_get_device("compass");
//  wb_compass_enable(comp, TIME_STEP);
  
  // init lidar
  wb_lidar_enable(lidar, TIME_STEP);
  lidar_width = wb_lidar_get_horizontal_resolution(lidar);
  int max_range = wb_lidar_get_max_range(lidar);
  printf("Lidar width = %d, lidar maxRanges = %d\n",lidar_width,max_range);
  wb_lidar_enable_point_cloud(lidar);
  printf("Sick width = %d\n",lidar_width);
  

  // Light Sensors
  leftLight = wb_robot_get_device("lsLeft");
  wb_light_sensor_enable(leftLight, TIME_STEP); 
  rightLight = wb_robot_get_device("lsRight");
  wb_light_sensor_enable(rightLight, TIME_STEP); 
  leftLightBack = wb_robot_get_device("lsLeftBack");
  wb_light_sensor_enable(leftLightBack, TIME_STEP); 
  rightLightBack = wb_robot_get_device("lsRightBack");
  wb_light_sensor_enable(rightLightBack, TIME_STEP); 
      
  // init the display
  wb_display_set_color(display, 0x0000ff);
  background =  wb_display_image_load(display, "grid_display.png");

  // init motors
  wb_motor_set_position(left_wheel, INFINITY);
  wb_motor_set_position(right_wheel, INFINITY);

  // Gripper motors
  gripper_motors[0] = wb_robot_get_device("lift motor");
  gripper_motors[1] = wb_robot_get_device("left finger motor");
  gripper_motors[2] = wb_robot_get_device("right finger motor");
  
  // Left and right primary ultrasonic sensors
  USleft3 = wb_robot_get_device("so3");
  USright4 = wb_robot_get_device("so4");
  wb_distance_sensor_enable(USleft3,TIME_STEP);
  wb_distance_sensor_enable(USright4,TIME_STEP);
     
  // Get the time step
  time_step = wb_robot_get_basic_time_step();
}

// This is a blocking timer and will let previous motor drive
// settings run for a number of seconds
void step(double seconds) {
  const double ms = seconds * 1000.0;
  int elapsed_time = 0;
  while (elapsed_time < ms) {
    wb_robot_step(time_step);
    elapsed_time += time_step;
  }
}

// This is a blocking stop and stops the robot wheels for a certain
// time
void stop(double seconds) {
  wb_motor_set_velocity(left_wheel, 0.0);
  wb_motor_set_velocity(right_wheel, 0.0);
  step(seconds);
}

// Sets gripper motors into lift (or fall)
void lift(double position) {
  wb_motor_set_velocity(gripper_motors[0], GRIPPER_MOTOR_MAX_SPEED);
  wb_motor_set_position(gripper_motors[0], position);
}

// Opens or closes gripper fingers
void moveFingers(double position) {
  wb_motor_set_velocity(gripper_motors[1], GRIPPER_MOTOR_MAX_SPEED);
  wb_motor_set_velocity(gripper_motors[2], GRIPPER_MOTOR_MAX_SPEED);
  wb_motor_set_position(gripper_motors[1], position);
  wb_motor_set_position(gripper_motors[2], position);
}

void turn(double speed) {
  wb_motor_set_velocity(left_wheel, speed);
  wb_motor_set_velocity(right_wheel, -speed);
}

void moveForwards(double speed) {
  wb_motor_set_velocity(left_wheel, speed);
  wb_motor_set_velocity(right_wheel, speed);
}

int main(int argc, char **argv) {

  double thetaObj;
  double omegaRobot;
  double omegaRobotLight; 
  double omegaRobotForce; 
  double forceRep;
  float range;
  int count;
  double omegaL,omegaR;
  int state;
  double velyRobot;
  bool breakFlag = false;
  
  wb_robot_init();

  init();
  
  state = STATE_FIND_OBJECT;
  
  // Grab Controller arguments
  k_att = atof(argv[1]);
  k_rep = atof(argv[2]);
  d_0   = atof(argv[3]);
  
  for(int i = 0;i<argc;i++)
    printf("argv[%d] = %s\n",i,argv[i]);
    
while (wb_robot_step(TIME_STEP) != -1) {
  
  // get lidar values
  const float *lidar_values = wb_lidar_get_range_image(lidar);

  update_display(display, background, lidar_values, lidar_width);

  
  switch(state) {
    case STATE_FIND_LIGHT:
      printf("\n");
      // get light values and set thetaRobot
      double leftIntensity  = wb_light_sensor_get_value(leftLight);
      double rightIntensity = wb_light_sensor_get_value(rightLight); 
      //printf("leftIntensity = %f\n",leftIntensity);

      omegaRobotLight = k_att*(leftIntensity - rightIntensity)/200;
      velyRobot = 0.5*MAX_SPEED;  
      
      if(leftIntensity > 195 || rightIntensity > 195){
        omegaRobotLight = 0;
        velyRobot = 0;
        state = STATE_PUTDOWN_OBJECT;
        printf("Transit to STATE_PUTDOWN_OBJECT\n");        
      }
           
      break;
      
    case STATE_FIND_OBJECT:   
      lift(0.05);        //lower the gripper
      moveFingers(0.1);  // Open its fingers
      step(0.5);         // Do above in 0.5 seconds
  
      count = 0;
      forceRep = 0;
      thetaObj = 0;
 
      velyRobot = 0.5*MAX_SPEED; 
      breakFlag = false;  
      for(int i = 0;i<lidar_width;i++) {
        range = lidar_values[i];     
        // transform angle and calc force
        if(range < 0.25) {
          breakFlag = true;
          //printf("BreakFlag set \n");
        } else if(range < RANGE_THRESHOLD) {
          thetaObj += (PI/2 - (double)i*PI/180);
          forceRep += compute_force_obstacle(range);
          count++;
        }
       } 
 
       double avTheta = thetaObj/(double)count ;
       if(count > 0 && !breakFlag) {
         printf("AvTheta = %f\n",avTheta);
         if(avTheta < 0) {
             omegaRobotForce = forceRep*(avTheta);
         } else if(avTheta > 0){
             omegaRobotForce = forceRep*(avTheta);
         } else {
           omegaRobot = 0;   
         }  
       }  
  
       if(breakFlag){
         velyRobot = 0;
         omegaRobotForce = 0;
         omegaRobotLight = 0;
         state = STATE_PICKUP_OBJECT;
         printf("Transit to STATE_PICKUP_OBJECT\n");
       }
       break;
  
  case STATE_PICKUP_OBJECT:
      printf("Close Gripper\n");
      moveFingers(0.02);     // Close fingers
      step(2.5);             // This is done in 2.5 secs     
      printf("Lift\n");
      lift(0.01);            // Lift the gripped object
      step(2.5);             // This takes 2.5 secs
      state = STATE_FIND_LIGHT;      
      printf("Transit to STATE_FIND_LIGHT\n");
  break;
  
  case STATE_PUTDOWN_OBJECT:
    lift(0.05);          // Put down the object
    step(2.0);           // This takes 2.0 secs
    moveFingers(0.1);    // Open the fingers
    step(2.0);           // This takes 2.0 secs
    state = STATE_DONE;
    printf("Transit to STATE_DONE\n");
    break;
    
  case STATE_DONE:
    stop(1.0);   
    break;
    
  }
  
  
    omegaRobot = omegaRobotForce + omegaRobotLight;
        
    omegaR = velyRobot + omegaRobot;
    omegaL = velyRobot - omegaRobot;

    wb_motor_set_velocity(left_wheel, omegaL);
    wb_motor_set_velocity(right_wheel, omegaR);
  }

  wb_robot_cleanup();

  return 0;
}
