// ====================================================================================
//
// CBP_Salamander_4
//
// 19-11-20 RELEASE
//
// Salamander procedural walking and swimming combined
//
// =====================================================================================

#include <assert.h>
#include <math.h>
#include <stdio.h>
#include <webots/distance_sensor.h>
#include <webots/gps.h>
#include <webots/keyboard.h>
#include <webots/motor.h>
#include <webots/robot.h>

// for array indexing
#define X 0
#define Y 1
#define Z 2

// 6 actuated body segments and 4 legs 
#define FIRST_SEG 0
#define LAST_SEG 5
#define FIRST_LEG 6
#define LAST_LEG 9

#define NUM_MOTORS 10
#define WATER_LEVEL 0.0
#define CONTROL_STEP 16

// locomotion modes

enum locomotion_modes {
  WALK,
  SWIM}
locomotion;

// State Variables
static double spine_offset = 0.0;
static double phase = 0.0; 
double b;
double L;

  // segments and legs in a single array for coding efficiency
  WbDeviceTag motor[NUM_MOTORS];
  double target_angle[NUM_MOTORS] = {0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0};

  // distance sensors (left and right eyes) and gps
  WbDeviceTag eyeL_tag, eyeR_tag, gps; 

  // get the motors device tags from the PROTO
  char *MOTOR_NAMES[NUM_MOTORS] = {"motor_1", "motor_2","motor_3","motor_4","motor_5",
                                   "motor_6", "motor_leg_1", "motor_leg_2", "motor_leg_3","motor_leg_4"};

// motors angle range 
static double min_motor_angle[NUM_MOTORS];
static double max_motor_angle[NUM_MOTORS];

double clamp(double value, double min, double max) {
  if (min > max) {
    assert(0);
    return value;
  } else if (min == 0 && max == 0) {
    return value;
  }

  return value < min ? min : value > max ? max : value;
}

void init() {
  // Get min and max allowed angles for all segments and legs
  for (int i = 0; i < NUM_MOTORS; i++) {
    motor[i] = wb_robot_get_device(MOTOR_NAMES[i]);
    min_motor_angle[i] = wb_motor_get_min_position(motor[i]);
    max_motor_angle[i] = wb_motor_get_max_position(motor[i]);
  }

  // get and enable left and right distance sensors 
  eyeL_tag = wb_robot_get_device("ds_left");
  wb_distance_sensor_enable(eyeL_tag, CONTROL_STEP);
  eyeR_tag = wb_robot_get_device("ds_right");
  wb_distance_sensor_enable(eyeR_tag, CONTROL_STEP);
  
  // get and enable gps device 
  gps = wb_robot_get_device("gps");
  wb_gps_enable(gps, CONTROL_STEP);

}


int main() {

  const double FREQUENCY = 1.4; // locomotion frequency [Hz] 
  const double WALK_AMPL = 0.3; // radians 
  const double SWIM_AMPL = 1.0; // radians 

  wb_robot_init();

  init();
 
  b = 0.5;
  L = 6.0;
  
  while (wb_robot_step(CONTROL_STEP) != -1) {

    // perform sensor measurment 
    double eyeL_val = wb_distance_sensor_get_value(eyeL_tag);
    double eyeR_val = wb_distance_sensor_get_value(eyeR_tag);

    // change direction according to sensor reading 
    spine_offset = (eyeR_val - eyeL_val);
    
    // get mode whether in or out of water
    double elevation = wb_gps_get_values(gps)[Y];
    if (locomotion == SWIM && elevation > WATER_LEVEL - 0.003) {
      locomotion = WALK;
      phase = target_angle[6];
    } else if (locomotion == WALK && elevation < WATER_LEVEL - 0.015) {
      locomotion = SWIM;   
    }
    
    phase -= (double)CONTROL_STEP / 1000.0 * FREQUENCY * 2.0 * M_PI;

    if(locomotion == WALK) {
    // Rotate segments
    for (int i = 0; i < 6; i++)
       target_angle[i] = WALK_AMPL * sin(phase) + spine_offset;
       
     //rotate legs
     target_angle[6] = phase;
     target_angle[7] = phase + M_PI;
     target_angle[8] = phase + M_PI;
     target_angle[9] = phase;

    
    } else if(locomotion == SWIM) {
    
       // put legs backwards for swimming 
    // fmod(x,y) returns remainder of x divided by y.
    double backwards_angle = phase - fmod(phase, 2.0 * M_PI) - M_PI / 2;
    for (int i = FIRST_LEG; i <= LAST_LEG; i++)
      target_angle[i] = backwards_angle;

    // below water level: swim (travelling wave of robot body) 
    for (int i = FIRST_SEG; i <= LAST_SEG; i++)
      target_angle[i] = SWIM_AMPL * sin(phase + i * (2 * M_PI / L)) * ((i/L + b) / (1 + b)) + spine_offset;
    }
    
       // Move all segments and all four legs
    for (int i = 0; i < NUM_MOTORS; i++) {
      target_angle[i] = clamp(target_angle[i], min_motor_angle[i], max_motor_angle[i]);
      wb_motor_set_position(motor[i], target_angle[i]);
    }
 
    
  }

  wb_robot_cleanup();

  return 0; 
}
